use reqwest::Method;

// SPDX-FileCopyrightText: 2022 Matteo Settenvini <matteo.settenvini@montecristosoftware.eu>
// SPDX-License-Identifier: AGPL-3.0-or-later

use {
    crate::api_client::ApiClient,
    crate::recipe::{Ingredient, Recipe},
    anyhow::{anyhow, Result},
    chrono::{Duration, Local},
    icalendar::Component,
    regex::Regex,
    reqwest::StatusCode,
    std::collections::HashSet,
    std::ops::Range,
};

pub async fn with(
    api_client: &ApiClient,
    calendar_name: &str,
    location: &str,
    days: u32,
) -> Result<()> {
    let ids = map_events_to_recipe_ids(api_client, calendar_name, days).await?;
    let ingredients = get_ingredients(api_client, ids).await?;
    let ingredients = merge_ingredients(ingredients);
    let md = prepare_grocery_list(&ingredients)?;
    // TODO filename is hardcoded for now
    save_grocery_list(api_client, location, &md).await?;
    Ok(())
}

async fn map_events_to_recipe_ids(
    api_client: &ApiClient,
    calendar_name: &str,
    days: u32,
) -> Result<HashSet<usize>> {
    let date_range = Range {
        start: Local::now(),
        end: Local::now() + Duration::days(days as i64),
    };

    let all_events = api_client.get_events(calendar_name, date_range).await?;

    let recipe_id_regex: Regex = Regex::new(r"cookbook@(\d+)").unwrap();
    let recipe_ids = all_events
        .iter()
        .flat_map(|event| event.property_value("DESCRIPTION"))
        .flat_map(|descr| recipe_id_regex.captures(descr))
        .flat_map(|c| c.get(1))
        .flat_map(|m| m.as_str().parse::<usize>())
        .collect::<HashSet<_>>();

    Ok(recipe_ids)
}

async fn get_ingredients<RecipeIds>(
    api_client: &ApiClient,
    recipe_ids: RecipeIds,
) -> Result<Vec<Ingredient>>
where
    RecipeIds: IntoIterator<Item = usize>,
{
    let ingredients = recipe_ids.into_iter().map(|id: usize| async move {
        // TODO code duplicated with schedule_csv::get_all_recipes
        let recipe_url = format!("apps/cookbook/api/recipes/{id}");
        let response = api_client
            .rest()
            .get(api_client.base_url().join(&recipe_url).unwrap())
            .send()
            .await
            .expect(&format!("Cannot fetch recipe with id {}", id));

        response.json::<Recipe>().await.map(|r| r.ingredients)
    });

    let ingredients = futures::future::try_join_all(ingredients).await?;
    Ok(ingredients.into_iter().flatten().collect())
}

fn merge_ingredients(mut ingredients: Vec<Ingredient>) -> Vec<Ingredient> {
    ingredients.sort();

    // TODO actual merging

    ingredients
}

fn prepare_grocery_list(ingredients: &Vec<Ingredient>) -> Result<String> {
    let mut out = String::new();
    use std::fmt::Write;

    writeln!(
        out,
        "# Grocery list - {}",
        chrono::Local::now().format("%Y-%m-%d").to_string()
    )?;
    writeln!(out)?; // leave an empty line
    for ingredient in ingredients {
        let ingredient = ingredient.0.as_str();
        writeln!(out, "- [ ] {}", ingredient)?;
    }

    Ok(out)
}

async fn save_grocery_list(api_client: &ApiClient, filename: &str, contents: &str) -> Result<()> {
    let dav_base_url = api_client
        .base_url()
        .join(&format!("remote.php/dav/files/{}/", api_client.username()))?;

    let filename_components = filename.split('/').collect::<Vec<_>>();
    filename_components
        .iter()
        .take(filename_components.len() - 1)
        .fold(Ok(dav_base_url.clone()), |url, dir| {
            url.map(|u| u.join(&format!("{dir}/")).unwrap())
                .and_then(|url| {
                    futures::executor::block_on(async {
                        let response = api_client
                            .rest()
                            .request(Method::from_bytes(b"MKCOL").unwrap(), url.clone())
                            .send()
                            .await;

                        match response.map(|r| r.status()) {
                            Ok(StatusCode::OK)
                            | Ok(StatusCode::METHOD_NOT_ALLOWED /* already exists */) => Ok(url),
                            Ok(status) => Err(anyhow!(
                                "Could not create WebDAV collection {}, server responded with {}",
                                &url,
                                status
                            )),
                            Err(e) => Err(anyhow!(e)),
                        }
                    })
                })
        })?;

    let file_url = dav_base_url.join(filename).unwrap();
    let response = api_client
        .rest()
        .put(file_url.clone())
        .header("Content-Type", "text/markdown; charset=utf-8")
        .body(contents.to_owned())
        .send()
        .await?;

    match response.status() {
        StatusCode::CREATED | StatusCode::NO_CONTENT => Ok(()),
        status => Err(anyhow!(
            "Cannot save grocery list at {}, server responded with status {}",
            file_url,
            status
        )),
    }
}
