// SPDX-FileCopyrightText: 2022 Matteo Settenvini <matteo.settenvini@montecristosoftware.eu>
// SPDX-License-Identifier: AGPL-3.0-or-later

use {
    crate::recipe::Recipe,
    chrono::{Datelike, NaiveDate, NaiveDateTime, NaiveTime, Utc},
    icalendar::Event as CalEvent,
    std::rc::Rc,
};

pub struct Event {
    pub uid: String,
    pub ends_at: NaiveDateTime,
    pub recipe: Rc<Recipe>,
}

#[derive(strum_macros::Display)]
pub enum Meal {
    Lunch,
    Dinner,
}

impl Event {
    pub fn new(date: NaiveDate, meal: Meal, recipe: Rc<Recipe>) -> Self {
        let uid = format!(
            "{}-{}@{}.montecristosoftware.eu",
            date,
            meal,
            env!("CARGO_PKG_NAME")
        );

        let meal_time = match meal {
            Meal::Lunch => NaiveTime::from_hms(12, 00, 00),
            Meal::Dinner => NaiveTime::from_hms(19, 00, 00),
        };

        let ends_at = NaiveDateTime::new(date, meal_time);

        Event {
            uid,
            ends_at,
            recipe,
        }
    }
}

impl From<Event> for CalEvent {
    fn from(ev: Event) -> Self {
        use icalendar::Component;

        const DAY_NAMES: [&str; 7] = ["MO", "TU", "WE", "TH", "FR", "SA", "SU"];
        let start_time = ev.ends_at - ev.recipe.total_time();

        let cal_event = CalEvent::new()
            .uid(&ev.uid)
            .summary(&ev.recipe.name)
            .description(&format!("cookbook@{}", ev.recipe.id))
            .location(&ev.recipe.url)
            .timestamp(Utc::now())
            .starts(start_time)
            .ends(ev.ends_at)
            .add_property(
                // TODO make configurable yearly repetition, for now this suits my personal uses
                "RRULE",
                &format!(
                    "FREQ=YEARLY;BYDAY={weekday};BYWEEKNO={weekno}",
                    weekday = DAY_NAMES
                        .get(start_time.weekday().num_days_from_monday() as usize)
                        .unwrap(),
                    weekno = start_time.iso_week().week(),
                ),
            )
            .done();

        cal_event
    }
}
